//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this sample source code is subject to the terms of the Microsoft
// license agreement under which you licensed this sample source code. If
// you did not accept the terms of the license agreement, you are not
// authorized to use this sample source code. For the terms of the license,
// please see the license agreement between you and Microsoft or, if applicable,
// see the LICENSE.RTF on your install media or the root of your tools installation.
// THE SAMPLE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
#include <windows.h>
#include <pnp.h>
#include <tchddi.h>
#include <tchstream.h>
#include <guts.h>

#include "../App/cetouchview.h"     // only for BUTTON_WIDTH, BUTTON_HEIGHT

static const DWORD MAX_DWORD = 0xFFFFFFFF;

// Comment this to stop sending touch data to GWES
#define ENABLE_GWES_TOUCH

///=============================================================================
/// This module implements the proxy interface to the touch driver.
///=============================================================================

/// Exported functions:
///     TouchPanelEnable()
///     TouchPanelDisable()
///     TouchPanelGetDeviceCaps()
///     TouchPanelSetMode()
///     TouchPanelPowerHandler()
///     TouchPanelSetCalibration()
///     TouchPanelReadCalibrationPoint()
///     TouchPanelReadCalibrationAbort()
///     TouchPanelCalibrateAPoint()

///=============================================================================
/// External functions
///=============================================================================

///=============================================================================
/// Local defines
///=============================================================================
static PFN_TOUCH_PANEL_GET_DEVICE_CAPS          s_pfnTouchPanelGetDeviceCaps;
static PFN_TOUCH_PANEL_SET_MODE                 s_pfnTouchPanelSetMode;
static PFN_TOUCH_PANEL_SET_CALIBRATION          s_pfnTouchPanelSetCalibration;
static PFN_TOUCH_PANEL_CALIBRATE_A_POINT        s_pfnTouchPanelCalibrateAPoint;
static PFN_TOUCH_PANEL_READ_CALIBRATION_POINT   s_pfnTouchPanelReadCalibrationPoint;
static PFN_TOUCH_PANEL_READ_CALIBRATION_ABORT   s_pfnTouchPanelReadCalibrationAbort;
static PFN_TOUCH_PANEL_POWER_HANDLER            s_pfnTouchPanelPowerHandler;
static PFN_TOUCH_PANEL_INITIALIZE_CURSOR        s_pfnTouchPanelInitializeCursor;
static PFN_TOUCH_PANEL_REGISTER_CALLBACK	 	s_pfnTouchPanelRegisterCallback;

static PFN_TOUCH_PANEL_ENABLE                   s_pfnTouchPanelEnable;
static PFN_TOUCH_PANEL_ENABLE_EX                s_pfnTouchPanelEnableEx;

static PFN_TOUCH_PANEL_CALLBACK    s_samplesCallBackFn   = NULL;
static PFN_TOUCH_PANEL_CALLBACK_EX s_samplesCallBackFnEx = NULL;

static HINSTANCE s_hProxy;
static HANDLE s_hMsgQ;

static const TCHAR sc_szProxyDll[] = _T("TchProxy.dll");

static FILE *s_fpMacro;
static BOOL s_fRecording = FALSE;
static BOOL s_fIgnoring = FALSE;
static DWORD s_dwPreviousTickCount = 0;
static RECT s_rcButton = {0};

DWORD WINAPI CETouchAutoTest(LPVOID);
void SendAutoTestData();


///=============================================================================
/// Local functions
///=============================================================================

LPWSTR StrDup ( LPWSTR lpstr )
{
    LPWSTR p;
    size_t len = 0;
    StringCchLength(lpstr, STRSAFE_MAX_CCH, &len);
    p = (LPWSTR)malloc((len + 1) * sizeof(WCHAR));
    if(p) {
        StringCchCopyN(p, len+1, lpstr, len+1);
    }
    return p;
}

// Call this immediately before or after calling m_samplesCallBackFnEx
// to send the points to the CETouchView application
static
void
BroadcastSamples(
    DWORD cInputs,
    PCETOUCHINPUT pInputs
    )
{
    static BOOL s_fQueueFull = FALSE;

    if (s_fRecording && (NULL != s_fpMacro))
    {
        // if ignoring input, check for an UP to restart recording
        if (    (s_fIgnoring) &&
                (1 == cInputs) &&
                (pInputs[0].dwFlags & TOUCHEVENTF_UP))
        {
            s_fIgnoring = FALSE;
        }
        // if not ignoring input, check for a DOWN in the button area to start ignoring until next UP
        else if (   (!s_fIgnoring) &&
                    (1 == cInputs) &&
                    (pInputs[0].dwFlags & TOUCHEVENTF_DOWN) &&
                    (pInputs[0].x >= s_rcButton.left*4) &&
                    (pInputs[0].x <= s_rcButton.right*4) &&
                    (pInputs[0].y >= s_rcButton.top*4) &&
                    (pInputs[0].y <= s_rcButton.bottom*4))
        {
            s_fIgnoring = TRUE;
        }
        // otherwise record the data if not ignoring
        else if (!s_fIgnoring)
        {
            // We're storing our own time deltas instead of what's in pInputs to reproduce the delays during playback...
            DWORD dwTickCount = GetTickCount();
            DWORD dwTime = dwTickCount - s_dwPreviousTickCount;
            s_dwPreviousTickCount = dwTickCount;

            switch (cInputs)
            {
                case 1:
                    fwprintf_s(s_fpMacro, L"%d,%d,%d,%d,%d,%d,%d,-1,-1,-1,-1,-1,-1\n",
                        dwTime,
                        pInputs[0].x,
                        pInputs[0].y,
                        pInputs[0].dwFlags,
                        pInputs[0].dwMask,
                        pInputs[0].cxContact,
                        pInputs[0].cyContact);
                    break;
                case 2:
                    fwprintf_s(s_fpMacro, L"%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d\n",
                        dwTime,
                        pInputs[0].x,
                        pInputs[0].y,
                        pInputs[0].dwFlags,
                        pInputs[0].dwMask,
                        pInputs[0].cxContact,
                        pInputs[0].cyContact,
                        pInputs[1].x,
                        pInputs[1].y,
                        pInputs[1].dwFlags,
                        pInputs[1].dwMask,
                        pInputs[1].cxContact,
                        pInputs[1].cyContact);
                    break;
                default:
                    ERRORMSG(1, (L"Error: cInputs is not 1 or 2!!!"));
                    ASSERT(FALSE);
            }
        }
    }

    BOOL fSuccess = WriteMsgQueue(s_hMsgQ, (LPVOID) pInputs, sizeof(CETOUCHINPUT) * cInputs, 0, 0);
    if (fSuccess)
    {
        s_fQueueFull = FALSE;
    }
    else
    {
        DWORD dwErr = GetLastError();

        if (ERROR_TIMEOUT != dwErr)
        {
            ERRORMSG(1, (L"WriteMsgQueue failed. Error=%u\r\n", dwErr));
        }
        else if (!s_fQueueFull)
        {
            RETAILMSG(1, (L"CETouchView message queue is now full\r\n"));
            s_fQueueFull = TRUE;
        }
    }
}


BOOL
CETouchFilter_Callback(
    DWORD cInputs,
    __in_ecount(cInputs) const PCETOUCHINPUT pInputs,
    DWORD cbSize
    )
{
    for (int i = 0; i < (int)cInputs; i++)
    {
        if (pInputs[i].dwTime == 0)
            pInputs[i].dwTime = GetTickCount();
		//RETAILMSG(1, (L"%S: [%d/%d] x=%d y=%d\r\n", i+1, cInputs, 
		//			__FUNCTION__, pInputs[i].x, pInputs[i].y));
    }

    BroadcastSamples(cInputs, pInputs);
#ifdef ENABLE_GWES_TOUCH
    return s_samplesCallBackFnEx(cInputs, pInputs, cbSize);
#else
    return TRUE;
#endif
}

BOOL GetButtonLocation ( RECT *pRC )
{
    HKEY hKeyTouch;
    DWORD dwType = REG_DWORD;
    DWORD dwSize = sizeof(DWORD);
    DWORD dwValue;

    SetRect(pRC,0,0,0,0);

    if (ERROR_SUCCESS != RegOpenKeyEx(HKEY_LOCAL_MACHINE, L"SOFTWARE\\Microsoft\\CETouchView", 0, 0, &hKeyTouch))
    {
        // should not happen
        return FALSE;
    }

    if (RegQueryValueEx(hKeyTouch, L"ButtonX", NULL, &dwType, (PBYTE)&dwValue, &dwSize)
            == ERROR_SUCCESS)
    {
        pRC->left = dwValue;
    }
    else
    {
        RegCloseKey(hKeyTouch);
        return FALSE;
    }

    if (RegQueryValueEx(hKeyTouch, L"ButtonY", NULL, &dwType, (PBYTE)&dwValue, &dwSize)
            == ERROR_SUCCESS)
    {
        pRC->top = dwValue;
    }
    else
    {
        pRC->left = 0;       // X is there but not Y???
        RegCloseKey(hKeyTouch);
        return FALSE;
    }
    RegCloseKey(hKeyTouch);

    pRC->right = pRC->left + BUTTON_WIDTH;
    pRC->bottom = pRC->top + BUTTON_HEIGHT;

    return TRUE;
}

void GetTouchDriverName(LPTSTR *ppDriverName)
{
    HKEY hKeyTouch;
    DWORD dwType = REG_SZ;
    DWORD dwSize = 128;
    TCHAR szDrvNameInRegistry[128];

    *ppDriverName = (LPTSTR)sc_szProxyDll;

    if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, L"HARDWARE\\DEVICEMAP\\TOUCH", 0, 0, &hKeyTouch)
            != ERROR_SUCCESS)
    {
        // should not happen
        return;
    }

    if (RegQueryValueEx(hKeyTouch, L"DriverExName", NULL, &dwType, (PBYTE)szDrvNameInRegistry, &dwSize)
            == ERROR_SUCCESS)
    {
        *ppDriverName = StrDup(szDrvNameInRegistry);
    }
    RegCloseKey(hKeyTouch);
}

BOOL GetRecordingMode ( void )
{
    HKEY hKeyTouch;
    DWORD dwType = REG_DWORD;
    DWORD dwSize = sizeof(DWORD);
    DWORD dwValue = 0;

    if (ERROR_SUCCESS != RegOpenKeyEx(HKEY_LOCAL_MACHINE, L"SOFTWARE\\Microsoft\\CETouchView", 0, 0, &hKeyTouch))
    {
        return FALSE;
    }

    RegQueryValueEx(hKeyTouch, L"MacroMode", NULL, &dwType, (PBYTE)&dwValue, &dwSize);
    RegCloseKey(hKeyTouch);

    return (0 != dwValue);
}

BOOL GetMacroName ( LPTSTR *ppMacroName )
{
    HKEY hKeyTouch;
    DWORD dwType = REG_SZ;
    DWORD dwSize = 128;
    TCHAR szMacroNameInRegistry[128];

    if (NULL != *ppMacroName)
    {
        free(*ppMacroName);
        *ppMacroName = NULL;
    }

    if (ERROR_SUCCESS != RegOpenKeyEx(HKEY_LOCAL_MACHINE, L"SOFTWARE\\Microsoft\\CETouchView", 0, 0, &hKeyTouch))
    {
        return FALSE;
    }

    if (ERROR_SUCCESS == RegQueryValueEx(hKeyTouch, L"FileName", NULL, &dwType, (PBYTE)szMacroNameInRegistry, &dwSize))
    {
        *ppMacroName = StrDup(szMacroNameInRegistry);
    }

    RegCloseKey(hKeyTouch);

    return ((NULL != *ppMacroName) && (L'\0' != *(*ppMacroName)));
}

///=============================================================================
/// Public functions
///=============================================================================

/* Register the callback function and start the callback thread if non-existent*/
BOOL
TouchPanelEnable(
    PFN_TOUCH_PANEL_CALLBACK    pfnCallback
    )
{
    ASSERT(FALSE);
    return FALSE; // This function should not be called since we expose the Ex version.
}

/* Register the callback function and start the callback thread if non-existent*/
BOOL
TouchPanelEnableEx(
    PFN_TOUCH_PANEL_CALLBACK_EX pfnCallbackEx
    )
{
    BOOL fRet = FALSE;
    LPTSTR pDriverName;

    MSGQUEUEOPTIONS opts = {
        sizeof(opts),
        MSGQUEUE_ALLOW_BROKEN,
        100,
        sizeof(CETOUCHINPUT[CETOUCHINPUT_MAX_SIMULTANEOUS]),
        FALSE
    };

    ASSERT(NULL == s_samplesCallBackFn); // Only called with one or the other
    s_samplesCallBackFnEx = pfnCallbackEx;

    GetTouchDriverName(&pDriverName);
    s_hProxy = LoadDriver(pDriverName);
    if (NULL == s_hProxy)
    {
        ERRORMSG(1, (L"Could not load the proxy driver (%s)\r\n", pDriverName));
        goto Error;
    }

    s_hMsgQ = CreateMsgQueue(L"TouchInput", &opts);
    if (NULL == s_hMsgQ)
    {
        ERRORMSG(1, (L"Could not create the message queue. Error=%u\r\n", GetLastError()));
        goto Error;
    }

#define LoadEntry(Name, Type)                                                   \
    s_pfn##Name = (PFN_##Type)GetProcAddress(s_hProxy,__TEXT(#Name));       \
    if ( s_pfn##Name == NULL )                                                  \
        {                                                                       \
        ERRORMSG(1,                                                             \
            (TEXT("Could not find ") __TEXT(#Name) __TEXT(".\r\n")));           \
        goto Error;                                               \
        }

    LoadEntry(TouchPanelGetDeviceCaps,          TOUCH_PANEL_GET_DEVICE_CAPS);
    LoadEntry(TouchPanelSetMode,                TOUCH_PANEL_SET_MODE);
    LoadEntry(TouchPanelSetCalibration,         TOUCH_PANEL_SET_CALIBRATION);
    LoadEntry(TouchPanelCalibrateAPoint,        TOUCH_PANEL_CALIBRATE_A_POINT);
    LoadEntry(TouchPanelReadCalibrationPoint,   TOUCH_PANEL_READ_CALIBRATION_POINT);
    LoadEntry(TouchPanelReadCalibrationAbort,   TOUCH_PANEL_READ_CALIBRATION_ABORT);
    LoadEntry(TouchPanelEnable,                 TOUCH_PANEL_ENABLE);
    LoadEntry(TouchPanelPowerHandler,           TOUCH_PANEL_POWER_HANDLER);
    LoadEntry(TouchPanelRegisterCallback,       TOUCH_PANEL_REGISTER_CALLBACK);

    LoadEntry(TouchPanelEnableEx,               TOUCH_PANEL_ENABLE_EX);
#undef LoadEntry

#define LoadEntry(Name, Type)   \
    s_pfn##Name = (PFN_##Type)GetProcAddress(s_hProxy,__TEXT(#Name));

    LoadEntry(TouchPanelInitializeCursor,       TOUCH_PANEL_INITIALIZE_CURSOR);
#undef LoadEntry

    fRet = s_pfnTouchPanelEnableEx(&CETouchFilter_Callback);

    CreateThread(NULL, 0, CETouchAutoTest, NULL, 0, NULL);

    s_fpMacro = NULL;
    s_fRecording = FALSE;

Error:
    return fRet;
}

BOOL
TouchPanelGetDeviceCaps(
    INT iIndex,
    LPVOID lpOutput
    )
{
    return s_pfnTouchPanelGetDeviceCaps(iIndex, lpOutput);
}

BOOL
TouchPanelSetMode(
    INT     iIndex,
    LPVOID  lpInput
    )
{
    return s_pfnTouchPanelSetMode(iIndex, lpInput);
}

void
TouchPanelPowerHandler(
    BOOL    bOff
    )
{
    s_pfnTouchPanelPowerHandler(bOff);
}

BOOL
TouchPanelSetCalibration(
    INT32   cCalibrationPoints,     //The number of calibration points
    INT32   *pScreenXBuffer,        //List of screen X coords displayed
    INT32   *pScreenYBuffer,        //List of screen Y coords displayed
    INT32   *pUncalXBuffer,         //List of X coords collected
    INT32   *pUncalYBuffer          //List of Y coords collected
    )
{
    return s_pfnTouchPanelSetCalibration(cCalibrationPoints, pScreenXBuffer, pScreenYBuffer, pUncalXBuffer, pUncalYBuffer);
}

BOOL
TouchPanelReadCalibrationPoint(
    INT *pRawX,
    INT *pRawY
    )
{
    return s_pfnTouchPanelReadCalibrationPoint(pRawX, pRawY);
}

VOID
TouchPanelReadCalibrationAbort(
    VOID
    )
{
    s_pfnTouchPanelReadCalibrationAbort();
}

VOID
TouchPanelCalibrateAPoint(
    INT32 UncalX,
    INT32 UncalY,
    INT32* pCalX,
    INT32* pCalY
    )
{
    s_pfnTouchPanelCalibrateAPoint(UncalX, UncalY, pCalX, pCalY);
}

PFN_TOUCH_PANEL_CALLBACK_EX
TouchPanelRegisterCallback(
    PFN_TOUCH_PANEL_CALLBACK_EX pfnCallbackEx
    )
{
    return s_pfnTouchPanelRegisterCallback(pfnCallbackEx);
}


BOOL WINAPI
DllMain(
    HANDLE hinstDll,
    DWORD dwReason,
    LPVOID lpRsvd
    )
{
    switch ( dwReason )
    {
        case DLL_PROCESS_ATTACH:
            DisableThreadLibraryCalls((HMODULE) hinstDll);
            break;
        case DLL_PROCESS_DETACH:
            break;
    }
    return ( TRUE );

}

DWORD WINAPI CETouchAutoTest(LPVOID)
{
    HANDLE s_hEvtAutoData = CreateEvent(NULL, FALSE, FALSE, L"CETouchAutoData");
    LPTSTR pMacroName = NULL;

    for (;;)
    {
        WaitForSingleObject(s_hEvtAutoData, INFINITE);

        // If we're currently in recording mode, this is our signal to stop
        if (s_fRecording)
        {
            fclose(s_fpMacro);
            s_fpMacro = NULL;
            s_fRecording = FALSE;
        }
        // Check registry value HKLM/??? if empty send auto test data, if set deal with named file
        else if (GetMacroName(&pMacroName))
        {
            // Now check for recording or playback mode
            s_fRecording = GetRecordingMode();

            if (s_fRecording)
            {
                errno_t iErr = _wfopen_s(&s_fpMacro, pMacroName, L"wt");
                if (0 != iErr)
                {
                    ERRORMSG(1, (L"Error (%d) creating file: \"%s\"", iErr, pMacroName));
                    free(pMacroName);
                    pMacroName = NULL;
                    s_fRecording = FALSE;
                }
                else
                {
                    s_dwPreviousTickCount = GetTickCount();           // time recording was started
                }
                if (!GetButtonLocation(&s_rcButton))
                {
                    ERRORMSG(1, (L"Unable to find button location!"));
                }
            }
            else
            {
                errno_t iErr = _wfopen_s(&s_fpMacro, pMacroName, L"rt");
                if (0 != iErr)
                {
                    ERRORMSG(1, (L"Error (%d) opening file: \"%s\"", iErr, pMacroName));
                    free(pMacroName);
                    pMacroName = NULL;
                }
                else
                {
                    CETOUCHINPUT rgInputs[2];
                    DWORD dwTime;

                    s_dwPreviousTickCount = GetTickCount();           // time playback was started
                    ZeroMemory(rgInputs, sizeof(rgInputs));
                    rgInputs[0].dwID = 0;
                    rgInputs[1].dwID = 1;
                    rgInputs[0].hSource = (HANDLE)1;
                    rgInputs[1].hSource = (HANDLE)1;
                    while (!feof(s_fpMacro))
                    {
                        WCHAR szInput[256];

                        if ((NULL != fgetws(szInput, ARRAYSIZE(szInput), s_fpMacro)) &&
                            (*szInput != L'\0') &&          // ignore blank lines...
                            (*szInput != L'\n') &&          //   this would mean blank too...
                            (*szInput != L';'))             //   and lines that begin with ';'
                        {
                            swscanf_s(szInput, L"%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d",
                                &dwTime,
                                &rgInputs[0].x, &rgInputs[0].y, &rgInputs[0].dwFlags, &rgInputs[0].dwMask, &rgInputs[0].cxContact, &rgInputs[0].cyContact,
                                &rgInputs[1].x, &rgInputs[1].y, &rgInputs[1].dwFlags, &rgInputs[1].dwMask, &rgInputs[1].cxContact, &rgInputs[1].cyContact);

                            if (INFINITE == dwTime) // Guard against infinite loop
                            {
                                dwTime = 0;
                            }

                            Sleep(dwTime);      // delay the appropriate time

                            int cInputs = (rgInputs[1].x == -1) ? 1 : 2;
                            CETouchFilter_Callback(cInputs, rgInputs, sizeof(CETOUCHINPUT) * cInputs);
                        }
                    }
                    fclose(s_fpMacro);
                }
            }
        }
        else
        {
            SendAutoTestData();
        }
    }

}


void SendAutoTestData()
{
    LONG x1, y1, x0, y0;
    DWORD dwScreenW = GetSystemMetrics(SM_CXSCREEN);
    DWORD dwScreenH = GetSystemMetrics(SM_CYSCREEN);

    CETOUCHINPUT rgInputs[CETOUCHINPUT_MAX_SIMULTANEOUS] = { 0 };

    x0 = dwScreenW / 2 - 30;
    y0 = dwScreenH / 2 - 30;
    rgInputs[0].dwID = 0;
    rgInputs[0].dwMask = TOUCHINPUTMASKF_CONTACTAREA;
    rgInputs[0].cxContact = 40;
    rgInputs[0].cyContact = 20;
    rgInputs[0].hSource = (HANDLE)1;

    x1 = dwScreenW / 2 + 30;
    y1 = dwScreenH / 2 + 30;
    rgInputs[1].dwID = 1;
    rgInputs[1].dwMask = TOUCHINPUTMASKF_CONTACTAREA;
    rgInputs[1].cxContact = 40;
    rgInputs[1].cyContact = 20;
    rgInputs[1].hSource = (HANDLE)1;

    rgInputs[0].dwFlags = TOUCHEVENTF_DOWN | TOUCHEVENTF_INRANGE | TOUCHEVENTF_PRIMARY | TOUCHEVENTF_CALIBRATED;
    rgInputs[1].dwFlags = TOUCHEVENTF_DOWN | TOUCHEVENTF_INRANGE | TOUCHEVENTF_CALIBRATED;

    rgInputs[0].x = x0 * 4, rgInputs[0].y = y0 * 4;
    rgInputs[1].x = x1 * 4, rgInputs[1].y = y1 * 4;
    CETouchFilter_Callback(2, rgInputs, sizeof(CETOUCHINPUT) * 2);
    Sleep(200);

    for (int k = 0; k < 2; k++)
    {
        for (unsigned int i = 0; i < 20; ++i)
        {
            Sleep(100);
            x0 -= 8;
            y0 -= 8;
            rgInputs[0].dwFlags = TOUCHEVENTF_MOVE | TOUCHEVENTF_INRANGE | TOUCHEVENTF_PRIMARY | TOUCHEVENTF_CALIBRATED;

            x1 += 8;
            y1 += 8;
            rgInputs[1].dwFlags = TOUCHEVENTF_MOVE | TOUCHEVENTF_INRANGE | TOUCHEVENTF_CALIBRATED;

            rgInputs[0].x = x0 * 4, rgInputs[0].y = y0 * 4;
            rgInputs[1].x = x1 * 4, rgInputs[1].y = y1 * 4;
            CETouchFilter_Callback(2, rgInputs, sizeof(CETOUCHINPUT) * 2);
        }

        for (i = 0; i < 20; ++i)
        {
            Sleep(100);
            x0 += 8;
            y0 += 8;
            rgInputs[0].dwFlags = TOUCHEVENTF_MOVE | TOUCHEVENTF_INRANGE | TOUCHEVENTF_PRIMARY | TOUCHEVENTF_CALIBRATED;

            x1 -= 8;
            y1 -= 8;
            rgInputs[1].dwFlags = TOUCHEVENTF_MOVE | TOUCHEVENTF_INRANGE | TOUCHEVENTF_CALIBRATED;

            rgInputs[0].x = x0 * 4, rgInputs[0].y = y0 * 4;
            rgInputs[1].x = x1 * 4, rgInputs[1].y = y1 * 4;
            CETouchFilter_Callback(2, rgInputs, sizeof(CETOUCHINPUT) * 2);
        }
    }

    rgInputs[0].dwFlags = TOUCHEVENTF_UP | TOUCHEVENTF_PRIMARY | TOUCHEVENTF_CALIBRATED;
    rgInputs[1].dwFlags = TOUCHEVENTF_UP | TOUCHEVENTF_CALIBRATED;

    rgInputs[0].x = x0 * 4, rgInputs[0].y = y0 * 4;
    rgInputs[1].x = x1 * 4, rgInputs[1].y = y1 * 4;
    CETouchFilter_Callback(2, rgInputs, sizeof(CETOUCHINPUT) * 2);

}
