//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this sample source code is subject to the terms of the Microsoft
// license agreement under which you licensed this sample source code. If
// you did not accept the terms of the license agreement, you are not
// authorized to use this sample source code. For the terms of the license,
// please see the license agreement between you and Microsoft or, if applicable,
// see the LICENSE.RTF on your install media or the root of your tools installation.
// THE SAMPLE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
//
//
//
// Copylight (C) 2011, UNITEC Co., LTD. All Rights Reserved.
// THIS SOURCE CODE, AND ITS USE AND DISTRIBUTION, IS SUBJECT TO THE TERMS
// AND CONDITIONS OF THE APPLICABLE LICENSE AGREEMENT
//

#include "utchid.h"
#include <devload.h>

#include "utc_ioctl.h"

#ifdef DEBUG

// Debug Zones
#define DBG_ERROR              0x0001
#define DBG_WARNING            0x0002
#define DBG_INIT               0x0004
#define DBG_FUNCTION           0x0008

#define DBG_USAGES             0x0010

DBGPARAM dpCurSettings = {
        TEXT("UtcHid"), {
        TEXT("Errors"), TEXT("Warnings"), TEXT("Init"), TEXT("Function"),
        TEXT("Usages"), TEXT(""), TEXT(""), TEXT(""),
        TEXT(""), TEXT(""), TEXT(""), TEXT(""),
        TEXT(""), TEXT(""), TEXT(""), TEXT("") },
        DBG_ERROR | DBG_WARNING };

static
void
ValidateHidUTC(
    PHID_UTC pHidUtc
    );

#else

#define ValidateHidUTC(ptr)

#endif // DEBUG

static VOID
FreeHidUTC(
    PHID_UTC pHidUtc
    );

// Create Message Queue for command response
static
BOOL
ResponseQueue_Create(
    PHID_UTC pHidUtc
    );

// Close Message Queue Handler
static
void
ResponseQueue_Close(
    PHID_UTC pHidUtc
    );

// Enqueue the command response
static
BOOL
ResponseQueue_Enqueue(
    PHID_UTC pHidUtc, 
    PCHAR pbBuffer, 
    DWORD cbBytes
    );

// Dequeue the command response
static
BOOL
ResponseQueue_Dequeue(
    PHID_UTC pHidUtc,
    PCHAR pbBuffer, 
    DWORD cbBufSize, 
    DWORD *pcbRecvSize
    );

// Clear the Command response
static
void
ResponseQueue_Clear(
    PHID_UTC pHidUtc
    );
    
// Dll entry function.
extern "C" 
BOOL
DllEntry(
    HANDLE hDllHandle,
    DWORD dwReason, 
    LPVOID lpReserved
    )
{
    UNREFERENCED_PARAMETER(lpReserved);
    
    switch (dwReason) {
        case DLL_PROCESS_ATTACH:
            DEBUGREGISTER((HINSTANCE)hDllHandle);

            DEBUGMSG(ZONE_INIT, (_T("%s: Attach\r\n"), __FUNCTION__));
            DisableThreadLibraryCalls((HMODULE) hDllHandle);
            break;
            
        case DLL_PROCESS_DETACH:
            //DEBUGMSG(ZONE_INIT, (_T("%s: Detach\r\n"), __FUNCTION__));
            break;
            
        default:
            break;
    }
    
    return TRUE;
}


// Get interrupt reports from the device. Thread exits when the device has
// been removed.
static
DWORD
WINAPI
UTCThreadProc(
    LPVOID lpParameter
    )
{
    PHID_UTC pHidUtc = (PHID_UTC) lpParameter;
    
    PREFAST_DEBUGCHK(pHidUtc != NULL);

    SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST);
    
    const DWORD cbMaxHidPacket = pHidUtc->hidpCaps.InputReportByteLength;
    PCHAR pbHidPacket = (PCHAR) LocalAlloc(LMEM_FIXED, cbMaxHidPacket);
    if (pbHidPacket == NULL) {
        DEBUGMSG(ZONE_ERROR, (TEXT("%s: LocalAlloc error:%d\r\n"), __FUNCTION__, GetLastError()));
        goto EXIT;
    }

    for (;;) {
        ValidateHidUTC(pHidUtc);

        DWORD cbHidPacket;
        // Get an interrupt report from the device.
        DWORD dwErr = pHidUtc->pHidFuncs->lpGetInterruptReport(
            pHidUtc->hDevice,
            pbHidPacket,
            cbMaxHidPacket,
            &cbHidPacket,
            pHidUtc->hevClosing,
            INFINITE);

        if (dwErr != ERROR_SUCCESS) {
            if ( (dwErr != ERROR_DEVICE_REMOVED) && (dwErr != ERROR_CANCELLED) ) {
                DEBUGMSG(ZONE_ERROR, 
                    (_T("%s: GetInterruptReport returned unexpected error %u\r\n"),
                    __FUNCTION__, dwErr));
            }
            // Exit thread
            break;
        }
        else {
            DEBUGCHK(cbHidPacket <= cbMaxHidPacket);

            if (!ResponseQueue_Enqueue(pHidUtc, pbHidPacket, cbHidPacket)) {
                DEBUGMSG(ZONE_ERROR, (_T("%s: Failed to enqueue the interrupt report\r\n"), __FUNCTION__));
            }
        }
    }

EXIT:
    if (pbHidPacket != NULL) LocalFree(pbHidPacket);
    DEBUGMSG(ZONE_FUNCTION, (_T("%s: Exiting thread\r\n"), __FUNCTION__));
    
    return 0;
}


// Entry point for the HID driver. Initializes the structures for this 
// keyboard and starts the thread that will receive interrupt reports.
extern "C"
BOOL
HIDDeviceAttach(
    HID_HANDLE                  hDevice, 
    PCHID_FUNCS                 pHidFuncs,
    const HID_DRIVER_SETTINGS * /*pDriverSettings*/,
    PHIDP_PREPARSED_DATA        phidpPreparsedData,
    PVOID                     * ppvNotifyParameter,
    DWORD                       /*dwUnused*/
    )
{
    BOOL fRet = FALSE;
//    size_t cbUsages;
    PHID_UTC pHidUtc;
    const TCHAR * const pszClientKey = CLIENT_REGKEY_SZ;

    DEBUGCHK(hDevice != NULL);
    DEBUGCHK(pHidFuncs != NULL);
    DEBUGCHK(phidpPreparsedData != NULL);
    DEBUGCHK(ppvNotifyParameter != NULL);

    // Allocate this keyboard's data structure and fill it.
    pHidUtc = (PHID_UTC) LocalAlloc(LPTR, sizeof(HID_UTC));
    if (pHidUtc == NULL) {
        DEBUGMSG(ZONE_ERROR, (_T("%s: LocalAlloc error:%d\r\n"), __FUNCTION__, GetLastError()));
        goto EXIT;
    }

    pHidUtc->dwSig              = HID_UTC_SIG;
    pHidUtc->hDevice            = hDevice;
    pHidUtc->pHidFuncs          = pHidFuncs;
    pHidUtc->phidpPreparsedData = phidpPreparsedData;
    pHidUtc->hThread            = NULL;
    pHidUtc->hevClosing         = NULL;
    pHidUtc->hMsgQueueWriter    = NULL;
    pHidUtc->hMsgQueueReader    = NULL;
	pHidUtc->dwOpenCount = 0;

    HidP_GetCaps(pHidUtc->phidpPreparsedData, &pHidUtc->hidpCaps);

    InitializeCriticalSection(&pHidUtc->csSingle);

    if (ResponseQueue_Create(pHidUtc) == FALSE) {
        DEBUGMSG(ZONE_ERROR, (_T("%s: ResponseQueue_Create error\r\n"), __FUNCTION__));
        goto EXIT;
    }

    pHidUtc->hevClosing = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (pHidUtc->hevClosing == NULL) {
        DEBUGMSG(ZONE_ERROR, (_T("%s: Error creating event: %d\r\n"),
            __FUNCTION__, GetLastError()));
        goto EXIT;
    }
    
    // Create the thread that will receive reports from this device
    pHidUtc->hThread = CreateThread(NULL, 0, UTCThreadProc, pHidUtc, 0, NULL);
    if (pHidUtc->hThread == NULL) {
        DEBUGMSG(ZONE_ERROR, (_T("%s: Failed creating keyboard thread\r\n"), 
            __FUNCTION__));
        goto EXIT;
    }

#ifdef DEBUG
    pHidUtc->fhThreadInited = TRUE;  // Set Thread inited flag for Debug
#endif

    // Create the UTC device
    pHidUtc->hOSDevice = ActivateDevice(pszClientKey, (DWORD) pHidUtc);
    if (pHidUtc->hOSDevice == NULL) {
        DEBUGMSG(ZONE_ERROR, (_T("%s: Error on ActivateDevice(%s)\r\n"), 
            __FUNCTION__, pszClientKey));
        goto EXIT;
    }

    *ppvNotifyParameter = pHidUtc;
    ValidateHidUTC(pHidUtc);
    fRet = TRUE;

EXIT:
    if ((fRet == FALSE) && (pHidUtc != NULL)) {
        // Initialization failed. Clean up any resources that did allocate.
        FreeHidUTC(pHidUtc);
    }
    return fRet;
}

// Entry point for the HID driver to give us notifications.
extern "C" 
BOOL 
WINAPI
HIDDeviceNotifications(
    DWORD  dwMsg,
    WPARAM wParam, // Message parameter
    PVOID  pvNotifyParameter
    )
{
    BOOL fRet = FALSE;
    PHID_UTC pHidUtc = (PHID_UTC) pvNotifyParameter;

    UNREFERENCED_PARAMETER(wParam);

    if (VALID_HID_UTC(pHidUtc) == FALSE) {
        DEBUGMSG(ZONE_ERROR, (_T("%s: Received invalid structure pointer\r\n"), __FUNCTION__));
        goto EXIT;
    }

    ValidateHidUTC(pHidUtc);
    
    switch(dwMsg) {
        case HID_CLOSE_DEVICE:
            DeactivateDevice(pHidUtc->hOSDevice);	// Call UTC_Deinit

            // Wait for UTC_Deinit to complete so we can free our context.
            FreeHidUTC(pHidUtc);

            fRet = TRUE;
            break;

        default:
            DEBUGMSG(ZONE_ERROR, (_T("%s: Unhandled message %u\r\n"), __FUNCTION__));
            break;
    };

EXIT:
    return fRet;
}


// Free the memory used by pHidUtc
void
FreeHidUTC(
    PHID_UTC pHidUtc
    )
{
    PREFAST_DEBUGCHK(pHidUtc != NULL);

    // Close thread if it has been created.
    if (pHidUtc->hThread != NULL) {
        DEBUGCHK(pHidUtc->hevClosing != NULL);
        SetEvent(pHidUtc->hevClosing);
        WaitForSingleObject(pHidUtc->hThread, INFINITE);
        CloseHandle(pHidUtc->hThread);
    }
    
    ResponseQueue_Clear(pHidUtc);
    ResponseQueue_Close(pHidUtc);
    
    if (pHidUtc->hevClosing != NULL) {
        CloseHandle(pHidUtc->hevClosing);
    }
    
    DeleteCriticalSection( &pHidUtc->csSingle );

    LocalFree(pHidUtc);
}

static
PHID_UTC
GetContextFromReg(
    LPCTSTR  pszActivePath
    )
{
    PHID_UTC pHidUtc = NULL;
    HKEY hKey;
    LONG lStatus;

    if (pszActivePath != NULL) 
    {
        //
        // open the registry and read out our context pointer
        // since Dev Mgr doesn't pass it in.
        //
        lStatus = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
                                pszActivePath,
                                0,
                                0,
                                &hKey);

        if (lStatus == ERROR_SUCCESS) 
        {
            DWORD dwVal;
            DWORD dwType = DEVLOAD_CLIENTINFO_VALTYPE;
            DWORD dwValLen = sizeof(dwVal);
            lStatus = RegQueryValueEx( hKey,
                                       DEVLOAD_CLIENTINFO_VALNAME,
                                       NULL,
                                       &dwType,
                                       (LPBYTE)(&dwVal),
                                       &dwValLen);

            if ((lStatus == ERROR_SUCCESS) && (dwType == REG_DWORD))
            {
                // check the signature
                pHidUtc = (PHID_UTC) dwVal;
                if (HID_UTC_SIG != pHidUtc->dwSig ) {
                    DEBUGMSG(ZONE_ERROR, (_T("%s: Invalid signature!!\r\n"), __FUNCTION__));
                    DEBUGCHK(FALSE);
                    pHidUtc = NULL;
                } else {
                    DEBUGMSG(ZONE_INIT, (_T("%s: ActivePath: %s\r\n"), __FUNCTION__, pszActivePath));
                 }
            }

            RegCloseKey(hKey);

        }
        else {
            DEBUGMSG(ZONE_ERROR, (_T("%s: Open ActivePath failed\n"), __FUNCTION__));
        }
    }

   return pHidUtc;
}


extern "C" 
DWORD
UTC_Init (
    DWORD dwCtx
    )
{
    LPCTSTR pszActivePath = (LPCTSTR) dwCtx; // HKLM\Drivers\Active\xx
    PHID_UTC pHidUtc = NULL;
    DWORD dwRet = 0;

    DEBUGCHK(pszActivePath != NULL);

    DEBUGMSG(ZONE_FUNCTION, (_T("+%s"), __FUNCTION__));

    pHidUtc = GetContextFromReg(pszActivePath);

    if (VALID_HID_UTC(pHidUtc) == TRUE) {
        dwRet = (DWORD) pHidUtc;
    }
    else {
        DEBUGCHK(FALSE); // Who called ActivateDevice with a bad value?
    }

    DEBUGMSG(ZONE_FUNCTION, (_T("-%s"), __FUNCTION__));
    
    return dwRet;
}


extern "C" 
BOOL
UTC_Deinit (DWORD dwCtx)
{
    DEBUGMSG(ZONE_FUNCTION, (_T("+%s"), __FUNCTION__));
    DEBUGMSG(ZONE_FUNCTION, (_T("-%s"), __FUNCTION__));

    return TRUE;
}

extern "C" 
BOOL
UTC_PreDeinit (DWORD /*dwCtx*/)
{
    // This function is necessary so that UTC_Deinit is called only 
    // after no threads are in UTC_Write, UTC_Read, and UTC_IOControl.

    DEBUGMSG(ZONE_FUNCTION, (_T("+%s"), __FUNCTION__));
    DEBUGMSG(ZONE_FUNCTION, (_T("-%s"), __FUNCTION__));

    return TRUE;
}

extern "C" 
DWORD
UTC_Open (DWORD dwCtx, DWORD /*dwAccMode*/, DWORD /*dwShrMode*/)
{
    PHID_UTC pHidUtc = (PHID_UTC) dwCtx;    
    DWORD dwRet = dwCtx;
    DWORD dwErr = ERROR_SUCCESS;
    
    DEBUGMSG(ZONE_FUNCTION, (_T("+%s"), __FUNCTION__));
    
    EnterCriticalSection( &pHidUtc->csSingle );
    
    // Check open count
    if( pHidUtc->dwOpenCount > 0 ) {
        DEBUGMSG(ZONE_ERROR, (_T("Multi-Open Error\r\n")));
        dwErr = ERROR_ACCESS_DENIED;
        dwRet = 0;
        goto EXIT;
    }

    if (VALID_HID_UTC(pHidUtc) == FALSE) {
        DEBUGMSG(ZONE_ERROR, (_T("%s: ERROR_FILE_NOT_FOUND\r\n"), __FUNCTION__));
        dwErr = ERROR_FILE_NOT_FOUND;
        dwRet = 0;
    }

    // increment open count
    pHidUtc->dwOpenCount++;
    
EXIT:
    LeaveCriticalSection( &pHidUtc->csSingle );
    SetLastError(dwErr);
    
    DEBUGMSG(ZONE_FUNCTION, (_T("-%s"), __FUNCTION__));
    
    return dwRet;
}

extern "C" 
BOOL
UTC_Close (DWORD dwCtx)
{
    PHID_UTC pHidUtc = (PHID_UTC) dwCtx;    
    BOOL fRet = FALSE;
    DWORD dwErr = ERROR_SUCCESS;

    DEBUGMSG(ZONE_FUNCTION, (_T("+%s"), __FUNCTION__));

    EnterCriticalSection( &pHidUtc->csSingle );

    if (VALID_HID_UTC(pHidUtc) == FALSE) {
        DEBUGMSG(ZONE_ERROR, (_T("%s: ERROR_INVALID_HANDLE\r\n"), __FUNCTION__));
        dwErr = ERROR_INVALID_HANDLE;
        goto EXIT;
    }

    // decrement open count
    pHidUtc->dwOpenCount--;

    fRet = TRUE;

EXIT:
    LeaveCriticalSection( &pHidUtc->csSingle );

    SetLastError(dwErr);

    DEBUGMSG(ZONE_FUNCTION, (_T("-%s"), __FUNCTION__));
    
    return fRet;
}

extern "C" 
BOOL
UTC_PreClose (DWORD /*dwCtx*/)
{
    // This function is necessary so that UTC_Close is called only 
    // after no threads are in UTC_Write, UTC_Read, and UTC_IOControl.

    DEBUGMSG(ZONE_FUNCTION, (_T("+%s"), __FUNCTION__));
    DEBUGMSG(ZONE_FUNCTION, (_T("-%s"), __FUNCTION__));

    return TRUE;
}

extern "C"
DWORD
UTC_Write(
    DWORD dwCtx,
    LPCVOID pBuffer,
    DWORD Count
    )
{
    DEBUGMSG(ZONE_FUNCTION, (_T("+%s"), __FUNCTION__));

    PHID_UTC pHidUtc = (PHID_UTC) dwCtx;

    DWORD dwRet = (DWORD)-1;
    DWORD dwErr = ERROR_SUCCESS;

    EnterCriticalSection( &pHidUtc->csSingle );

    if (VALID_HID_UTC(pHidUtc) == FALSE) {
        DEBUGMSG(ZONE_ERROR, (_T("%s: ERROR_INVALID_HANDLE\r\n"), __FUNCTION__));
        dwErr = ERROR_INVALID_HANDLE;
        goto EXIT;
    }

    if (pBuffer == NULL){
        DEBUGMSG(ZONE_ERROR, (_T("%s: Input buffer is NULL\r\n"), __FUNCTION__));
        dwErr = ERROR_INVALID_PARAMETER;
        goto EXIT;
    }

    if(Count == 0){
        DEBUGMSG(ZONE_ERROR, (_T("%s: Input buffer size is zero\r\n"), __FUNCTION__));
        dwErr = ERROR_INVALID_PARAMETER;
        goto EXIT;
    }

    // Check output buffer length
    if(Count > pHidUtc->hidpCaps.OutputReportByteLength){
        DEBUGMSG(ZONE_ERROR, (_T("%s: Input buffer size is bigger than max response size\r\n"), __FUNCTION__));
        dwErr = ERROR_INVALID_PARAMETER;
        goto EXIT;
    }

    dwErr = ERROR_SUCCESS;
    dwErr = pHidUtc->pHidFuncs->lpIssueInterruptTransfer(pHidUtc->hDevice, (PCHAR)pBuffer, Count);

    if (dwErr != ERROR_SUCCESS) {
        if ( (dwErr != ERROR_DEVICE_REMOVED) && (dwErr != ERROR_CANCELLED) ) {
            DEBUGMSG(ZONE_ERROR, 
                (_T("%s: IssueInterruptTransfer returned unexpected error %u\r\n"),
                __FUNCTION__, dwErr));
        }
    }
    else {
        // Return the number of bytes written
        dwRet = Count;
    }

EXIT:
    LeaveCriticalSection( &pHidUtc->csSingle );

    SetLastError(dwErr);

    DEBUGMSG(ZONE_FUNCTION, (_T("-%s"), __FUNCTION__));

    return dwRet;
}


extern "C"
DWORD
UTC_Read(
    DWORD dwCtx,
    LPCVOID pBuffer,
    DWORD Count
    )
{
    DEBUGMSG(ZONE_FUNCTION, (_T("+%s"), __FUNCTION__));

    DWORD dwRet = (DWORD)-1;
    DWORD dwSize;
    PHID_UTC pHidUtc = (PHID_UTC) dwCtx;
    DWORD dwErr = ERROR_SUCCESS;

    EnterCriticalSection( &pHidUtc->csSingle );

    if (VALID_HID_UTC(pHidUtc) == FALSE) {
        DEBUGMSG(ZONE_ERROR, (_T("%s: ERROR_INVALID_HANDLE\r\n"), __FUNCTION__));
        dwErr = ERROR_INVALID_HANDLE;
        goto EXIT;
    }

    if (pBuffer == NULL){
        DEBUGMSG(ZONE_ERROR, (_T("%s: Input buffer is NULL\r\n"), __FUNCTION__));
        dwErr = ERROR_INVALID_PARAMETER;
        goto EXIT;
    }

    // Check input buffer length
    if(Count < pHidUtc->hidpCaps.InputReportByteLength){
        DEBUGMSG(ZONE_ERROR, (_T("%s: Input buffer size is less than max response size\r\n"), __FUNCTION__));
        dwErr = ERROR_INVALID_PARAMETER;
        goto EXIT;
    }
    
    // Get Pacekt from Msg Queue
    if(!ResponseQueue_Dequeue(pHidUtc, (PCHAR)pBuffer, Count, &dwSize)){
        DEBUGMSG(ZONE_ERROR, (_T("%s: ResponseQueue_Enqueue Error\r\n"), __FUNCTION__));
    }
    else {
        // Return the number of bytes read
        dwRet = dwSize;
    }

EXIT:
    LeaveCriticalSection( &pHidUtc->csSingle );
    
    SetLastError(dwErr);
 
    DEBUGMSG(ZONE_FUNCTION, (_T("-%s"), __FUNCTION__));

    return dwRet;
}


extern "C" 
BOOL
UTC_IOControl(
    DWORD dwCtx,
    DWORD dwCode,
    PBYTE pBufIn,
    DWORD dwLenIn,
    PBYTE /*pBufOut*/,
    DWORD /*dwLenOut*/,
    PDWORD /*pdwActualOut*/
    )
{
    PHID_UTC pHidUtc = (PHID_UTC) dwCtx;
    
    DWORD dwErr = ERROR_SUCCESS;

    DEBUGMSG(ZONE_FUNCTION, (_T("+%s"), __FUNCTION__));

    EnterCriticalSection( &pHidUtc->csSingle );

    switch (dwCode) {
    case IOCTL_UTC_GET_MAX_RESPONSE_LEN:
        if ((pBufIn == NULL) || (dwLenIn < sizeof(DWORD))) {
            DEBUGMSG(ZONE_ERROR, (_T("%s: Invalid Parameter IOCTL Code:0x%08X\r\n"), __FUNCTION__, GetLastError()));
            dwErr = ERROR_INVALID_PARAMETER;
        }
        else {
            // Set Max Response Length
            *((DWORD *)pBufIn) = pHidUtc->hidpCaps.InputReportByteLength;
        }
        break;
        
    case IOCTL_UTC_GET_MAX_COMMAND_LEN:
        if ((pBufIn == NULL) || (dwLenIn < sizeof(DWORD))) {
            DEBUGMSG(ZONE_ERROR, (_T("%s: Invalid Parameter IOCTL Code:0x%08X\r\n"), __FUNCTION__, GetLastError()));
            dwErr = ERROR_INVALID_PARAMETER;
        }
        else {
            // Set Max Command Length
            *((DWORD *)pBufIn) = pHidUtc->hidpCaps.OutputReportByteLength;
        }
        break;
        
    default:
        dwErr = ERROR_INVALID_CATEGORY;
        break;
    }

    LeaveCriticalSection( &pHidUtc->csSingle );

    SetLastError(dwErr);
    DEBUGMSG(ZONE_FUNCTION, (_T("-%s"), __FUNCTION__));
    return (dwErr == ERROR_SUCCESS);
}


#ifdef DEBUG

// Validate a PHID_UTC structure
static
void
ValidateHidUTC(
    PHID_UTC pHidUtc
    )
{    
    PREFAST_DEBUGCHK(pHidUtc != NULL);
    DEBUGCHK(pHidUtc->dwSig == HID_UTC_SIG);
    DEBUGCHK(pHidUtc->hDevice != NULL);
    DEBUGCHK(pHidUtc->pHidFuncs != NULL);
    DEBUGCHK(pHidUtc->phidpPreparsedData != NULL);
    DEBUGCHK(pHidUtc->hidpCaps.UsagePage == 0xFFFF/*HID USAGE PAGE = Vendor Define*/);
    DEBUGCHK(pHidUtc->hidpCaps.Usage == 0x1/*HID USAGE = 0x1*/);
    if (pHidUtc->fhThreadInited == TRUE) {
        DEBUGCHK(pHidUtc->hThread != NULL);
    }
    DEBUGCHK(pHidUtc->hevClosing != NULL);
    DEBUGCHK(pHidUtc->hMsgQueueWriter != NULL);
    DEBUGCHK(pHidUtc->hMsgQueueReader != NULL);
}

#endif // DEBUG

// Create Message Queue for command response
static
BOOL
ResponseQueue_Create(
    PHID_UTC pHidUtc
    )
{
    BOOL bRet;
    HANDLE hRet;
    MSGQUEUEOPTIONS reader_opts = {
        sizeof(reader_opts),
        0,
        UTCHID_MSG_QUEUE_MAX,
        sizeof( UTC_RESPONSE ),
        TRUE
    };
    MSGQUEUEOPTIONS writer_opts = {
        sizeof(writer_opts),
        0,
        UTCHID_MSG_QUEUE_MAX,
        sizeof( UTC_RESPONSE ),
        FALSE
    };
    
    DEBUGCHK(pHidUtc != NULL);
    DEBUGCHK(pHidUtc->hMsgQueueReader == NULL);
    DEBUGCHK(pHidUtc->hMsgQueueWriter == NULL);
    
    bRet = FALSE;
    
    hRet = CreateMsgQueue(UTCHID_MSG_QUEUE_NAME, &reader_opts);
    if (hRet == NULL)
    {
        DEBUGMSG(ZONE_ERROR, (_T("%s: Could not create the message queue for reader. Error=%u\r\n"), 
            __FUNCTION__, GetLastError()));
        goto EXIT;
    }
    pHidUtc->hMsgQueueReader = hRet;
    

    hRet = CreateMsgQueue(UTCHID_MSG_QUEUE_NAME, &writer_opts);
    if (hRet == NULL)
    {
        DEBUGMSG(ZONE_ERROR, (_T("%s: Could not create the message queue for writer. Error=%u\r\n"), 
            __FUNCTION__, GetLastError()));
        goto EXIT;
    }
    pHidUtc->hMsgQueueWriter = hRet;

    bRet = TRUE;

EXIT:

    return bRet;
}

// Close Message Queue Handler
static
void
ResponseQueue_Close(
    PHID_UTC pHidUtc
    )
{
    DEBUGCHK(pHidUtc != NULL);
    
    if (pHidUtc->hMsgQueueReader != NULL) {
        if (CloseMsgQueue(pHidUtc->hMsgQueueReader) == FALSE) {
            DEBUGMSG(ZONE_ERROR, (_T("%s: Could not close the message queue handle for reader. Error=%u\r\n"), 
                __FUNCTION__, GetLastError()));
        }
    }
    
    if (pHidUtc->hMsgQueueWriter != NULL) {
        if (CloseMsgQueue(pHidUtc->hMsgQueueWriter) == FALSE) {
            DEBUGMSG(ZONE_ERROR, (_T("%s: Could not close the message queue handle for writer. Error=%u\r\n"), 
                __FUNCTION__, GetLastError()));
        }
    }
}

// Enqueue the command response
static
BOOL
ResponseQueue_Enqueue(
    PHID_UTC pHidUtc, 
    PCHAR pBuffer, 
    DWORD dwBytes
    )
{
    UTC_RESPONSE utcResponse;
    BOOL bRet;

    bRet = FALSE;
    
    DEBUGCHK(pHidUtc != NULL);
    DEBUGCHK(pHidUtc->hMsgQueueWriter != NULL);
    
    // Allocate response buffer memory
    utcResponse.pBuffer = (PBYTE)LocalAlloc(LPTR, dwBytes);
    if (utcResponse.pBuffer == NULL) {
        DEBUGMSG(ZONE_ERROR, (_T("%s: LocalAlloc error:%d\r\n"), __FUNCTION__, GetLastError()));
        goto EXIT;
    }

    CopyMemory(utcResponse.pBuffer, pBuffer, dwBytes);
    utcResponse.dwBytes = dwBytes;

    if (!WriteMsgQueue(pHidUtc->hMsgQueueWriter, (LPVOID)&utcResponse, sizeof(UTC_RESPONSE), 0, 0)) {
        DWORD dwErr = GetLastError();
        if (ERROR_TIMEOUT != dwErr) {
            DEBUGMSG(ZONE_ERROR, (_T("WriteMsgQueue: failed. Error=%u\r\n"), dwErr));
        }
        else {
            DEBUGMSG(ZONE_ERROR, (_T("WriteMsgQueue: message queue is now full\r\n")));
        }
        bRet = FALSE;
    }
    else {
        bRet = TRUE;
    }
EXIT:
    if (bRet == FALSE) {
        if (utcResponse.pBuffer == NULL) {
            LocalFree(utcResponse.pBuffer);
        }
    }

    return bRet;
}

// Dequeue the command response
static
BOOL
ResponseQueue_Dequeue(
    PHID_UTC pHidUtc,
    PCHAR pbBuffer,
    DWORD cbBufSize,
    DWORD *pcbRecvSize
    )
{
    BOOL bRet;
    DWORD dwRead;
    DWORD dwFlags;
    UTC_RESPONSE utcResponse;

    DEBUGCHK(pHidUtc != NULL);
    DEBUGCHK(pHidUtc->hMsgQueueReader != NULL);

    bRet = FALSE;
    *pcbRecvSize = 0;

    if (!ReadMsgQueue(pHidUtc->hMsgQueueReader, &utcResponse, sizeof(utcResponse), &dwRead, 0, &dwFlags)) {
        DEBUGMSG(ZONE_ERROR, (_T("%s: ReadMsgQueue: failed. Error=%u\r\n"), __FUNCTION__, GetLastError()));
        goto EXIT;
    }
    
    if (dwRead != sizeof(utcResponse)) {
        DEBUGMSG(ZONE_ERROR, (_T("%s: Msg length mismatch\r\n"), __FUNCTION__));
        goto EXIT;
    }
    
    if (utcResponse.dwBytes > cbBufSize) {
        DEBUGMSG(ZONE_ERROR, (_T("%s: Response msg size is bigger than buffer size\r\n"), __FUNCTION__));
        goto EXIT;
    }
    
    CopyMemory(pbBuffer, utcResponse.pBuffer, utcResponse.dwBytes);
    LocalFree(utcResponse.pBuffer);
    *pcbRecvSize = utcResponse.dwBytes;
    
    bRet = TRUE;
EXIT:
    return bRet;
}

// Clear the Command response
static
void
ResponseQueue_Clear(
    PHID_UTC pHidUtc
    )
{
    UTC_RESPONSE utcResponse;
    DWORD dwRead;
    DWORD dwFlags;
    
    DEBUGCHK(pHidUtc != NULL);
    DEBUGCHK(pHidUtc->hMsgQueueReader != NULL);

    while(WaitForSingleObject(pHidUtc->hMsgQueueReader, 0) == WAIT_OBJECT_0) {
        if (!ReadMsgQueue(pHidUtc->hMsgQueueReader, &utcResponse, sizeof(utcResponse), &dwRead, 0, &dwFlags)) {
            DEBUGMSG(ZONE_ERROR, (_T("%s: ReadMsgQueue: failed. Error=%u\r\n"), __FUNCTION__, GetLastError()));
            continue;
        }
    
        if (dwRead != sizeof(utcResponse)) {
            DEBUGMSG(ZONE_ERROR, (_T("%s: Msg length mismatch\r\n"), __FUNCTION__));
            continue;
        }
    
        LocalFree(utcResponse.pBuffer);

        DEBUGMSG(ZONE_WARNING, (_T("%s: ReadMsgQueue Dummy Read\r\n"), __FUNCTION__));
    }
}